/* debuginfod utilities for GDB.
   Copyright (C) 2020-2021 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include "defs.h"
#include <errno.h>
#include "cli/cli-style.h"
#include "gdbsupport/scoped_fd.h"
#include "debuginfod-support.h"

#ifndef HAVE_LIBDEBUGINFOD
scoped_fd
debuginfod_source_query (const unsigned char *build_id,
			 int build_id_len,
			 const char *srcpath,
			 gdb::unique_xmalloc_ptr<char> *destname)
{
  return scoped_fd (-ENOSYS);
}

scoped_fd
debuginfod_debuginfo_query (const unsigned char *build_id,
			    int build_id_len,
			    const char *filename,
			    gdb::unique_xmalloc_ptr<char> *destname)
{
  return scoped_fd (-ENOSYS);
}

scoped_fd
debuginfod_exec_query (const unsigned char *build_id,
		       int build_id_len,
		       const char *filename,
		       gdb::unique_xmalloc_ptr<char> *destname)
{
  return scoped_fd (-ENOSYS);
}

int
debuginfod_set_soname_buildid (struct bfd *bfd, const bfd_build_id *build_id)
{
  return 0;
}

int
debuginfod_get_soname_buildid (char *soname, gdb::unique_xmalloc_ptr<char> *build_id_hex)
{
  return 0;
}

void
debuginfod_clear_soname_buildids ()
{
  return;
}
#else
#include "elf-bfd.h"
#include "gdbcore.h"
#include <libgen.h>
#include <elfutils/debuginfod.h>
#include <unordered_map>

static std::unordered_map<std::string, std::string> soname_to_buildid;

struct user_data
{
  user_data (const char *desc, const char *fname)
    : desc (desc), fname (fname), has_printed (false)
  { }

  const char * const desc;
  const char * const fname;
  bool has_printed;
};

static int
progressfn (debuginfod_client *c, long cur, long total)
{
  user_data *data = static_cast<user_data *> (debuginfod_get_user_data (c));

  if (check_quit_flag ())
    {
      printf_filtered ("Cancelling download of %s %ps...\n",
		       data->desc,
		       styled_string (file_name_style.style (), data->fname));
      return 1;
    }

  if (!data->has_printed && total != 0)
    {
      /* Print this message only once.  */
      data->has_printed = true;
      printf_filtered ("Downloading %s %ps...\n",
		       data->desc,
		       styled_string (file_name_style.style (), data->fname));
    }

  return 0;
}

static debuginfod_client *
debuginfod_init ()
{
  debuginfod_client *c = debuginfod_begin ();

  if (c != nullptr)
    debuginfod_set_progressfn (c, progressfn);

  return c;
}

/* See debuginfod-support.h  */

scoped_fd
debuginfod_source_query (const unsigned char *build_id,
			 int build_id_len,
			 const char *srcpath,
			 gdb::unique_xmalloc_ptr<char> *destname)
{
  if (getenv (DEBUGINFOD_URLS_ENV_VAR) == NULL)
    return scoped_fd (-ENOSYS);

  debuginfod_client *c = debuginfod_init ();

  if (c == nullptr)
    return scoped_fd (-ENOMEM);

  user_data data ("source file", srcpath);

  debuginfod_set_user_data (c, &data);
  scoped_fd fd (debuginfod_find_source (c,
					build_id,
					build_id_len,
					srcpath,
					nullptr));

  /* TODO: Add 'set debug debuginfod' command to control when error messages are shown.  */
  if (fd.get () < 0 && fd.get () != -ENOENT)
    printf_filtered (_("Download failed: %s.  Continuing without source file %ps.\n"),
		     safe_strerror (-fd.get ()),
		     styled_string (file_name_style.style (),  srcpath));
  else
    destname->reset (xstrdup (srcpath));

  debuginfod_end (c);
  return fd;
}

/* See debuginfod-support.h  */

scoped_fd
debuginfod_debuginfo_query (const unsigned char *build_id,
			    int build_id_len,
			    const char *filename,
			    gdb::unique_xmalloc_ptr<char> *destname)
{
  if (getenv (DEBUGINFOD_URLS_ENV_VAR) == NULL)
    return scoped_fd (-ENOSYS);

  debuginfod_client *c = debuginfod_init ();

  if (c == nullptr)
    return scoped_fd (-ENOMEM);

  char *dname = nullptr;
  user_data data ("separate debug info for", filename);

  debuginfod_set_user_data (c, &data);
  scoped_fd fd (debuginfod_find_debuginfo (c, build_id, build_id_len, &dname));

  if (fd.get () < 0 && fd.get () != -ENOENT)
    printf_filtered (_("Download failed: %s.  Continuing without debug info for %ps.\n"),
		     safe_strerror (-fd.get ()),
		     styled_string (file_name_style.style (),  filename));

  destname->reset (dname);
  debuginfod_end (c);
  return fd;
}

/* See debuginfod-support.h  */

scoped_fd
debuginfod_exec_query (const unsigned char *build_id,
                      int build_id_len,
                      const char *filename,
                      gdb::unique_xmalloc_ptr<char> *destname)
{
  const char *urls_env_var = getenv (DEBUGINFOD_URLS_ENV_VAR);
  if (urls_env_var == NULL || urls_env_var[0] == '\0')
    return scoped_fd (-ENOSYS);

  debuginfod_client *c = debuginfod_init ();

  if (c == nullptr)
    return scoped_fd (-ENOMEM);

  char *dname = nullptr;
  user_data data ("executable for", filename);

  debuginfod_set_user_data (c, &data);
  scoped_fd fd (debuginfod_find_executable (c, build_id, build_id_len, &dname));
  debuginfod_set_user_data (c, nullptr);

  if (fd.get () < 0 && fd.get () != -ENOENT)
    printf_filtered (_("Download failed: %s.  Continuing without executable for %ps.\n"),
                    safe_strerror (-fd.get ()),
                    styled_string (file_name_style.style (),  filename));

  if (fd.get () >= 0)
    destname->reset (dname);

  return fd;
}

/* Attempt to locate the DT_SONAME tag within ABFD's .dynamic section.
   If found copy the string referred to by the tag to SONAME and return 1.
   Otherwise return 0.  */

static int
scan_soname_tag (struct bfd *abfd, std::string &soname)
{
  if (abfd == nullptr)
    return 0;

  if (bfd_get_flavour (abfd) != bfd_target_elf_flavour)
    return 0;

  int arch_size = bfd_get_arch_size (abfd);
  if (arch_size == -1)
    return 0;

  /* Find the start address of the .dynamic section.  */
  struct bfd_section *sect = bfd_get_section_by_name (abfd, ".dynamic");
  if (sect == nullptr)
    return 0;

  /* Read in .dynamic from the BFD.  */
  int sect_size = bfd_section_size (sect);
  gdb::byte_vector bufstart (sect_size);

  if (!bfd_get_section_contents (abfd, sect, bufstart.data (), 0, sect_size))
    return 0;

  gdb_byte *buf = bufstart.data ();
  struct bfd_section *dynstr = nullptr;
  CORE_ADDR soname_idx = 0;
  int step = (arch_size == 32) ? sizeof (Elf32_External_Dyn)
                              : sizeof (Elf64_External_Dyn);

  /* Iterate over BUF and search for DT_SONAME and DT_STRTAB.  */
  for (gdb_byte *bufend = buf + sect_size;
       (buf < bufend) && (soname_idx == 0 || dynstr == nullptr);
       buf += step)
  {
    long current_dyntag;
    CORE_ADDR dyn_ptr;

    if (arch_size == 32)
      {
       Elf32_External_Dyn *x_dynp_32 = (Elf32_External_Dyn *) buf;
       current_dyntag = bfd_h_get_32 (abfd, (bfd_byte *) x_dynp_32->d_tag);
       dyn_ptr = bfd_h_get_32 (abfd, (bfd_byte *) x_dynp_32->d_un.d_ptr);
      }
    else
      {
       Elf64_External_Dyn *x_dynp_64 = (Elf64_External_Dyn *) buf;
       current_dyntag = bfd_h_get_64 (abfd, (bfd_byte *) x_dynp_64->d_tag);
       dyn_ptr = bfd_h_get_64 (abfd, (bfd_byte *) x_dynp_64->d_un.d_ptr);
      }

    if (current_dyntag == DT_NULL)
      buf = bufend;
    else if (current_dyntag == DT_SONAME)
      soname_idx = dyn_ptr;
    else if (current_dyntag == DT_STRTAB)
      for (struct bfd_section *s = abfd->sections; s != nullptr; s = s->next)
       if (s->vma == dyn_ptr)
         dynstr = s;
  }

  if (soname_idx == 0 || dynstr == nullptr)
    return 0;

  sect_size = bfd_section_size (dynstr);
  if (sect_size <= soname_idx)
    return 0;

  /* Read the soname from the string table.  */
  gdb::char_vector dynstr_buf (sect_size);
  if (!bfd_get_section_contents (abfd, dynstr, dynstr_buf.data (), 0, sect_size))
    return 0;

  soname = dynstr_buf.data () + soname_idx;
  return 1;
}

/* See debuginfod-support.h  */

int
debuginfod_set_soname_buildid (struct bfd *bfd, const bfd_build_id *build_id)
{
  if (bfd == nullptr || build_id == nullptr)
    return 0;

  gdb_bfd_ref_ptr abfd (gdb_bfd_open (bfd->filename, gnutarget));

  if (abfd == nullptr)
    return 0;

  /* Check that bfd is an ET_DYN ELF file.  */
  bfd_check_format (abfd.get (), bfd_object);
  if (bfd_get_flavour (abfd.get ()) != bfd_target_elf_flavour
      || !(bfd_get_file_flags (abfd.get ()) & DYNAMIC))
    return 0;

  std::string soname;
  std::string build_id_hex;

  /* Determine soname of shared library.  If found map soname to build-id.  */
  if (scan_soname_tag (abfd.get (), soname))
   {
      for (int i = 0; i < build_id->size; i++)
       {
         build_id_hex += "0123456789abcdef"[build_id->data[i] >> 4];
         build_id_hex += "0123456789abcdef"[build_id->data[i] & 0xf];
       }

      soname_to_buildid[soname] = build_id_hex;
      return 1;
    }

  return 0;
}

/* See debuginfod-support.h  */

int
debuginfod_get_soname_buildid (char *soname, gdb::unique_xmalloc_ptr<char> *build_id_hex)
{
  auto it = soname_to_buildid.find (basename (soname));
  if (it == soname_to_buildid.end ())
    return 0;

  build_id_hex->reset (xstrdup (it->second.c_str ()));
  return 1;
}

/* See debuginfod-support.h  */

void debuginfod_clear_soname_buildids ()
{
  soname_to_buildid.clear ();
  return;
}

#endif
